"""
Created on Thu Jan 20 01:25:44 2022

@author: Haotian Teng
"""

# from bwapy import BwaAligner
# from bwapy.libbwa import Alignment
import sys
import re
import pysam
import numpy as np
import mappy as mp
from numpy import ndarray
from typing import List, Optional
from xron.utils.seq_op import raw2seq
from itertools import permutations
from collections import defaultdict
from Bio.Seq import Seq

def reverse_complement(seq):    
    return str(Seq(seq).reverse_complement())
           
class MetricAligner(mp.Aligner):
    
    def __new__(cls,
                reference:str,
                min_len: int = 7,
                options:str = "map-ont",
                *args, **kw):
        """
        The wrapper aligner class of bwa aligner, measure the quality for 
        batch of output sequences from a basecall.

        Parameters
        ----------
        reference : str
            File path of a reference fasta/fastq file.
        min_len : int, optional
            The minimum sequence length to be aligned. The default is 7.
        options: str, the options to mappy, default is map-ont preset

        """
        return super().__new__(cls,fn_idx_in = reference,preset = options,*args, **kw)

    
    def __init__(self,
                 reference:str,
                 min_len: int = 7,
                 options:str = "map-ont",
                 *args, **kw):
        super().__init__()
        self.reference = {}
        with pysam.FastxFile(reference) as f:
            for entry in f:
                self.reference[entry.name] = entry.sequence

    def permute_align(self,
                      raw: ndarray,
                      alphabet:str = "NACGT",
                      mod_base:dict = {},
                      min_len:int = 5,
                      permute:bool = True
                      )->ndarray:
        """
        Align the input sequence with best permutation.

        Parameters
        ----------
        raw : A [N,L] ndarray gives the raw ctc sequence.
            where N is the number of sequences and L is the length of each
            sequence, where 0 alaways mean the blank symbol.
        permute: List[str] Optional
            A permute list, default is None, will try every possible permutation.
        mod_base: Dict Optional
            A dict mapping the modification base to the original base.
        Returns
        -------
        ndarray
            A arra with same length of the sequences, gives the identity scores.

        """
        
        ab = alphabet[1:]
        seqs = raw2seq(raw)
        len_score = np.asarray([len(seq) for seq in seqs])-min_len
        len_score[len_score<0] = 0
        scores = []
        perms = []
        if permute:
            for perm in permutations(ab):
                perm_seqs = ["".join([perm[x-1] for x in seq]) for seq in seqs]
                identities = self.align(perm_seqs)
                identities = identities * len_score
                scores.append(identities)
                perms.append(perm)
        else:
            seqs = ["".join([ab[x-1] for x in seq]) for seq in seqs]
            identities = self.align(seqs)
            identities = identities * len_score
            scores.append(identities)
            perms.append([x for x in ab])
        return scores, perms
        
        
    def align(self,
              sequences: List[str]
              ) -> ndarray:
        """
        Get the align score given a batch of sequences.

        Parameters
        ----------
        sequences : List[str]
            A list of the sequences.
        Returns
        -------
        ndarray
            A array with same length of the sequences, gives the identity
            scores.

        """
        identities = []
        for seq in sequences:
            hits = self.align_seq(seq)
            identity = self._match_score(hits[0])/len(seq) if len(hits)>0 else 0  #Only use the first hit.
            identities.append(identity)
        return np.array(identities)
    
    def align_seq(self,
                  seq: str)-> mp.Alignment:
        return list(self.map(seq))
    
    def _match_score(self, alignment: mp.Alignment) -> int:
        """
        Get the match score given a cigar string.
    
        Parameters
        ----------
        alignment : mp.Alignment
            A alignment instance from the bwapy mapping result.
    
        Returns
        -------
        int
            The nubmer of matched bases.
    
        """
        cigar_s = alignment.cigar_str
        n_match = self._count_cigar(cigar_s,'M')
        return n_match - alignment.NM

    def ref_seq(self,seq:str,exclude_negative = False):
        """
        Return the reference sequence of the first hit of the alignment
        An illustration of the ref_idx:
            The alignment is
            Rerference: - - A A C C C - G T T C
            Query:      C C A A - - C G G A T C
            query_idx:  0 1 2 3     4 5 6 7 8 9
            ref_seq:    A A C C C G T T C
            ref_idx:    2 3 3 4 4 6 7 8 9
        
        Parameters
        ----------
        seq : str
            The query sequence.

        Returns
        -------
        ref_seq: Related reference sequence 
        ref_idx: Index reference (index relatively to the query/basecalled sequence), equal length of ref_seq.

        """
        hits = self.align_seq(seq)
        if not hits:
            return [],None,None,None
        hit = hits[0]
        if exclude_negative and not hit.strand:
            return [],None,None,None
        cigar = hit.cigar_str
        n_deletion = self._count_cigar(cigar,'D')
        n_match = self._count_cigar(cigar,'M')
        ref_len = n_match + n_deletion 
        ref_seq = self.reference[hit.ctg][hit.r_st:hit.r_st+ref_len]
        ref_idx = np.zeros(len(ref_seq))
        increment = [int(x) for x in re.findall(r'(\d+)',cigar)]
        operation = re.findall(r'[A-Za-z]',cigar)
        curr = 0 #The current operation index of reference mapping index
        curr_query = hit.q_st #The current pointer to the query sequence
        quality_score = np.zeros(len(ref_seq))
        for inc,op in zip(increment,operation):
            if op == "S":
                curr_query += inc
            elif op == "M":
                ref_idx[curr:curr+inc] = np.arange(curr_query,curr_query + inc)
                quality_score[curr:curr+inc] = 1
                curr += inc
                curr_query += inc
            elif op == "I":
                quality_score[curr]-= inc
                curr_query += inc
            elif op == "D":
                if inc <= 1:
                    ref_idx[curr:curr+inc] = curr_query
                    quality_score[curr:curr+inc] = 0
                else:
                    ref_idx[curr:curr+inc//2] = curr_query-1
                    ref_idx[curr+inc//2:curr+inc] = curr_query
                    quality_score[curr:curr+inc] = 0
                curr += inc
        if hit.strand == -1:
            ref_seq = reverse_complement(ref_seq)
            ref_idx = len(seq) - ref_idx
            ref_idx = ref_idx[::-1]
            quality_score = quality_score[::-1]
        return hits,ref_seq,ref_idx,quality_score
                

    def _count_cigar(self,cigar_string,c = 'M'):
        return sum([int(x) for x in re.findall(r'(\d+)%s'%(c),cigar_string)])
    
if __name__ == "__main__":
    reference_f = '/home/heavens/bridge_scratch/NA12878_RNA_IVT/GRCh38_transcript/GRCh38_latest_rna.fna'
    # test_seqs = ['ACCTGGAACGATTGGTTTGCGAATCGTCGTCATTTATTAATTGGTTCTGGGAAGGATTCATTTTGGCTCTTCGGTCCGCCGGGTCCCTACCTCGACTTCCTCCACGTCTTAGGGGATCGGATCTTGCCAGTCCGCATCATCCACGCAACTCCGGTAAGTTGCTCCAGAACCAGCAGGCATCCTTCCCAAGTTCCGAGGTTAGGTATCGGTACCCGCCAGGTTCAATCGAATGCAATATACGTTAACGTTTGTCATAGAAACATCTATTAATTCCGCATGTTACTCACATTGACCACACACAGCTGGGTAAAGAGCAGAGGGCTCCTGTCAACATACAGGATGGCTTGATAACATTGGCTTGTCCTTTAATCTAAATATGAAAACCAATCAATACAGCAGCAAGGCCTGACACCGCAAGCCATGCCGCGTGGTGACGTGCTATCCCAATTGCACGCTGGAT',
    #             'AUAGAUUGUGUUUGUUAGUCGCUUGGUACGCCGAGCUGAAUGAUUGUGGAUAUCGCGAAGCAGGUUGAGCGCAACCCUGGACGACCGUAUUUCGGCGUCAAUAAGCGUCGUGAAGGCAGGUGAAUAUGAGCAUGGCGUUGGGGCUUUCCGAUAUCAAAGGCGCAGAUGAGCGGUCUGGUAUACCGUGGUUUGUGACUGGCAACGGCUUUAACGUUGGCGAUACGCUGCGUUAUCUGCAGGCAAUUGAGCGGCUGGAGAAAAACUGGCUGUCGAUCCGUCCGAACCGGGCGCGAAAUGCCGAAAGUCGAGCGUCCAGCAGGCGUGGCAUAAUGGAUAAUAAAACUGCCCAGCGCGUCGCGAAGAUGAUGAUGUUAAAGAGAUCCGUUUGGAUGAUUGAAGAACUGCGUGUCAGUAUUCGCCCAACGGUUACGGUGCCUUAUCCUAUCUCCGAUAAGCGUAUUUUACAGGCCAUGGAUCAGAUUACGGCCUAAAGCCAGGCGACGAUGAGCUAAAGAGAUCCAGUACAAGUUCGGAACAAAAAGGAAAUUAUGAAUACAACGCCAUUCCAGCGCUUUAAUGAAACCAUUGCUGGCGGUGAAUACACCCAAUGACAGUGGCUGGCGCAGCGUGAUUUUUCCGGCGAACAACCAUCAUCGGAUUCGCGACCGGCGGGCAAUGCGGUGAUCCCACAUCGACGCGGCCUGUCGCAUCGCCAAGCGCAGGCGUUCCACUGAUUGCGGCGGUAUAAUGGGUCACGCCUUCGACGCCGUUUUUGCGCCGUUGAUUGCCCGGCAUCCCGCGCUACCACACUAUCUGCCAACCGGACGCGCUGAAGCCAUCGAUCUCGCGGAUCGCGAACCAGUCUGGCAUAUUCCGGCUGAGAAAUCUGGCUGGAAGGUCGGUCAACUAUUGCGGCGAAAAUGCCCGUCACCUGCGUCAAUCCGCCAGGCGAAAGAAAACAUUAACGGCUAUCGUUGCGAAGACCCCACCAUGCAGCAGCGCACUAUCGCGUUAUUCCGGCGCGUGACAAACGACGACACCGAUGCGUGCUGGCUGAGUUUUUCCUGGUUUGUUCCAGUACUGCGCCACCUUAACGGCGAUACACGUUUUGCCUGUGUUGAAGAGGGCAUGGGACGGUGGAGCGUUAUCUGUACCGAGAUUGCCGGCGAGCUGCCUGCGUGAUGAAACAGGGUAAGCACCGCGCGGUAAAGAUUUUAUUUCACGUUGGAUAUCCCCACACGCGAUUAUUAGAAACGCAUUGGCGGUAUUAGCAGGCAGAUACGACGCUGCGCAGUGCAAUUAGAACAGAGAGCGUUCGCGAAAAAAUCCCGUCUGCGGGUGAACGGGCAAGUGCGACCACCAUAUCUCAUAAAUUUAAUGAAUAUUUAUAAAAGCAAAAUCGUUGUCGCUCACGGUUUCAUUCAGGACGCGCUAUGGGCGGCAAGUAUUCCGGCCUGCAAAUUGGUAUUUUACUGGUUAAGU',
    #             'AGACUUGUGUUUGGCGGUCGCUUCUAAACCAGCGUAAAAGA',
    #             'CAGAUAGAUAGUGUUUGUUAGUCGCUUUGGUGCGAAACUCCACUUAUCUACUUACUACGGUCAAAUUAACGAUCAAGCUUUAAGUACGUCGCCAGAUUGUCUGGUUUAACCCUCAAGUGCGGGACAAUUAUUGGGACGAAUACUACGAGUACGGUCGUGUGGCCAUCUCCACCGUCCCCCGCUGUAAUGCUUUUUUUACUCGUCUUAAGGAUAGGUAUCAUCUGACUAGUACCGGCGUGCAAAAGCUUUUUCCAGGAAAGUAGUCAGGAUGAGGCUGGUGAACUGUUACGAAAGAUGAAGGUAAGACCUCAUCCUACAUUUCACAGGCAAGCCAUUAAAGUGCACUGUCUCGUGGCGAUUGCAACUUCGAUCGGCAUCUGAUCAUUGGCCAGCCCUUAGCCGUGUUUCGGUGUGAAGCAGCGUGGCGGGAAUCGCCAUCAUGAUCCCCGCAUGUAGGGCCCUAUCGGAAUUACACCGUCCUAAUGGUUUUUGAAUUGCGGUGGCGCCAGAAAAGA',
    #             'AGAUUGUGUUUGUUAGUUGCUUAACAAAAUGGUACAGAGAUGAUAGAAAGGCCUGGAGCCGUCUUCCGGGAAUCCAGAACAGAAUCAUCAACAUUUCCACGAAUCUCAAAGAGCUUAUUGAUUUUGGGCUGAGGCAGCUUGGAAGAAAGCAAUCUCCCAGGCAAAAGCAGCUGUUUUUUUAUCCGUCGAACGGUACUGCACUGCUGUCAUGGGCCGACAGGUUCCAGUAGCUCCUUACCAUGCGAACCGGAUCUACCAGUUUGCCAUCUACUCUGACAUUUUGAAGCUCAGGUGCACCAUUUAUUACACUAACUGUGCAGAUUACCAAUCCAAACUGCUUUCAAUUCGGACACAAAAAAAAGGUUUUACAGGUGCCAGCGAGGAUAAGGCUGGUUUAGUCGAACAAGCCAAUGGAGCAUACACUCUUAUGGACGGCUCCGCCUCCCAAAGGCAAGAAAUGUUGUCUUGCUACUGACAGCGGUUCACCACAACGGCUUGGAGAAUCCUGAACAUAAACGAACAUCAGACGUAUUGUUUAUUUGUGCCACUACAGAAAACCUAGGUCUGCAACUGAAAACAUUUUUGCAGAAUCCCAAUGACCAUUUUAACCAUGUUGUAGAACGAUCGCUUAAAGAAAAGGGUGGAUCUGACGACGUUUUAUUUAGGAAAGAAGCAGAACGAAUUAAGAAAAAUCUGAGUGUGCACAUAGAUGUCUAUAAUGCGCUUAUUCAUUCUGCGAAAUUUGAAAUGUGGGACAGCUGAAAUCAAAUGUCCAGUUAGUUUGUGCACACAGGGUUUUUGUUCCAACCUUGACGGACCGAGGUUAAUGAAAUGACUGUUCGGGAUCUGCCGGAUGAAAUCAAGCAAGAAUGGAUGUCCAGCAGCAAAAUACUGCAAAGGAGCAAAGCCAUUUGAAUGUCAUAAUUUUUUCGAUCAUCUCCGAUUAAGCGGAAGAUGAAACAACCGAAUAUUGACGAGGACCCUUUUCAUUUAACUUGUAUCAUAUUGCGAUCGAAGAAAAGGAG',
    #             'AGAUUGUGUUUGUUAGUUGCUUAACAAAAUGGUACAGAGAUGAUAGAAAGGCCUGGAGCCGUCUUCCGGGAAUCCAGAACAGAAUCAUCAACAUUUCCACGAAUCUCAAAGAGCUUAUUGAUUUUGGGCUGAGGCAGCUUGGAAGAAAGCAAUCUCCCAGGCAAAAGCAGCUGUUUUUUUAUCCGUCGAACGGUACUGCACUGCUGUCAUGGGCCGACAGGUUCCAGUAGCUCCUUACCAUGCGAACCGGAUCUACCAGUUUGCCAUCUACUCUGACAUUUUGAAGCUCAGGUGCACCAUUUAUUACACUAACUGUGCAGAUUACCAAUCCAAACUGCUUUCAAUUCGGACACAAAAAAAAGGUUUUACAGGUGCCAGCGAGGAUAAGGCUGGUUUAGUCGAACAAGCCAAUGGAGCAUACACUCUUAUGGACGGCUCCGCCUCCCAAAGGCAAGAAAUGUUGUCUUGCUACUGACAGCGGUUCACCACAACGGCUUGGAGAAUCCUGAACAUAAACGAACAUCAGACGUAUUGUUUAUUUGUGCCACUACAGAAAACCUAGGUCUGCAACUGAAAACAUUUUUGCAGAAUCCCAAUGACCAUUUUAACCAUGUUGUAGAACGAUCGCUUAAAGAAAAGGGUGGAUCUGACGACGUUUUAUUUAGGAAAGAAGCAGAACGAAUUAAGAAAAAUCUGAGUGUGCACAUAGAUGUCUAUAAUGCGCUUAUUCAUUCUGCGAAAUUUGAAAUGUGGGACAGCUGAAAUCAAAUGUCCAGUUAGUUUGUGCACACAGGGUUUUUGUUCCAACCUUGACGGACCGAGGUUAAUGAAAUGACUGUUCGGGAUCUGCCGGAUGAAAUCAAGCAAGAAUGGAUGUCCAGCAGCAAAAUACUGCAAAGGAGCAAAGCCAUUUGAAUGUCAUAAUUUUUUCGAUCAUCUCCGAUUAAGCGGAAGAUGAAACAACCGAAUAUUGACGAGGACCCUUUUCAUUUAACUUGUAUCAUAUUGCGAUCGAAGAAAAGGAG',
    #             'AAAGTAAGCCTGCTCGCTAAGTAGTCTTGCCCGCTCGGTGCGCAGTTCAATCGCTTTCAACAG']
    test_seqs = ['AGCUGAUAUUGCUGGGGGGCGUUCGCAGCCGCCAUCCAGCGGAGGAGCGCAGCCCAGGAGGCCGGAGGACGCGCCCAUAGAAUGCCCAGGGCGACAAACUGCACCUGAGCCCCACUGGUGCAGCCACCCUGCCUGUCCCAUACGCCCCGCCCACACAUGGAGUCCAGAGGAAGUCAGCCAGCAGCCCAAGCCCGACCUUCCAGGUGCCCCAGGCCACCUCGAGGCCAGGUACCCCCGGCAGCAGCCGAGGGAAAGCCCCUUGACCAAGCCCAAAGAAGGAGGCCCCGGCCGAGAAGCAGCAGCCGCCAGCAGCCCCACCACGGCACCUGCCAAGAAGACUCGGCCAGGGCCGACCCUGCCCUUCAACAACCACAGCAACCUGAAGCCCAGCCCCCACGGUCCCCAAACAUAGUCCCGAUGCAACCCCGGAGCCCAAGGGUCCUGGGACGGGGCGGAGGAAGAUGAGGCUGCCAGUGGGGCCUGGGGGCCGAGGACCCCUGGUCCUGUGAGAACUCAACCCCUGCUGGUGGCUGGGGUGUGGCCGUGGCAGCCAUAGCCCCUGAUUCUCGGUGUGGCCUCCUGGUCCGGAAAAAUAAUACCUGGGCCAGGCGGGGGCACGGAGCCACUUCCUGUACAGACCUCCCGAGGAAGCCAGUGCACUAGAGUUUCACCCUUACCUAUUCGUACACGCACAUUCUACACCCUCACACAUGCCCCCAACACACACGCGCACGGUGAAGGACGCCCGAGCCCACCCCUGCUGACCCAGGACUUCCCCAACCCAGGGCAGAAAGAGCCCAGGCUCCGGGGUCCAGCACAGGAUGUGGGGCGAGGGCACAGCUGGGAACGGCAAGAAAGGAAUGGACCCUGGCGAAAAGGCCCCCACCCCCUGGCGGCUGGGUGGAUCGGGGCCCCCAGGGCUGGUCUGAGUGCAGGUGGGUGCCGGGAUGGGUUAGGCCUGGGCCCGGCCCUCGUGGGGACAUUAAAGGGCGCAGUGGCUCCACUCGCCCCAUCUCUGCUUUGCGUACCCACCGGGGGUCUGCCUGGGCGAAUGCAGUGGGACGGACCUGGGGUGGGCACUUCCUCAGCAGAGCCUCAAUGCCACCUCCCCACCUUUCACCCAAGCCUAGCUGGGUGGGAAGACGGAGCCCUGAAGCCUGCGUGUGGGGUGCUGCCCAAGGGUCGCCCUUCCCCAGCUUGUGCUCCAGGGAGCCCCGACGCCGGCGGGAGGCAGGGAUGUCGAGGGCCCCACAGCUGUGCCCGCCCCUGCCCCGCUAGACUGCCCUCCAGCUUGCACCCAGCACCUGAGUCAACCGGCGUCGUAAUGCAAUAACAGGUGGAGUAGAACUGCUUCUGGCGGCGACAUGCCACCUCUGUCCUCAGCCCACUGCACCUGGGAAGCCCCUCGUGCAGCCCUCCUAGCACAUCCAGUCCCAGCGUGGCCCAUGCGCGCAGCAGCAGGCGAACCCCGGAGGUGGUGGUGGCACUGUGAGAACGUGGCCUGCCCUCUGGCCAGUGGCCUGCAGGAACCCAGGCCUGAGUGAGGCUGAGUCUGAAAUAAACUCUGUCGUCUGAAAAAAAAAAAAAUAAAAAAAAAGAAGAUAGAGCGACAGGCAAGUAAAAG']
    aln = MetricAligner(reference_f)
    scores = aln.align(test_seqs)
    hits = aln.align_seq(test_seqs[0]) #The position is the postiive start position in reference genome even for reverse complement alignment.
    print(scores)
    print(hits)
    hits,ref_seq,ref_idx,quality_score = aln.ref_seq(test_seqs[0])
    # rev_seq = reverse_complement(aln.reference['bsubtilis1'][:100]+'ATACACACA')
    # aln.ref_seq(rev_seq)
    


"""
Created on Sun Feb 21 05:22:32 2021

@author: Haotian Teng
"""
# from bwapy import BwaAligner
# from bwapy.libbwa import Alignment
import sys
if 'bwapy' in sys.modules:
    raise ImportError("Loading bwapy again will cause error, please restart " \
                      "the process.")
import bwapy
import re
import numpy as np
from numpy import ndarray
from typing import List, Optional
from xron.utils.seq_op import raw2seq
from itertools import permutations

class MetricAligner(bwapy.BwaAligner):
    def __init__(self,
                 reference: str,
                 min_len: int = 7,
                 options:str = None
                 ):
        """
        The wrapper aligner class of bwa aligner, measure the quality for 
        batch of output sequences from a basecall.

        Parameters
        ----------
        reference : str
            File path of a reference fasta/fastq file.
        min_len : int, optional
            The minimum sequence length to be aligned. The default is 7.

        """
        if not options:
            options = '-x ont2d'
            # options = '-A 1 -B 1 -k %d -O 3 -T 0'%(min_len)
        super().__init__(index = reference,options = options)
    
    def permute_align(self,
                      raw: ndarray,
                      alphabet:str = "NACGT",
                      mod_base:dict = {},
                      min_len:int = 5,
                      permute:bool = True
                      )->ndarray:
        """
        Align the input sequence with best permutation.

        Parameters
        ----------
        raw : A [N,L] ndarray gives the raw ctc sequence.
            where N is the number of sequences and L is the length of each
            sequence, where 0 alaways mean the blank symbol.
        permute: List[str] Optional
            A permute list, default is None, will try every possible permutation.
        mod_base: Dict Optional
            A dict mapping the modification base to the original base.
        Returns
        -------
        ndarray
            A arra with same length of the sequences, gives the identity scores.

        """
        
        ab = alphabet[1:]
        seqs = raw2seq(raw)
        len_score = np.asarray([len(seq) for seq in seqs])-min_len
        len_score[len_score<0] = 0
        scores = []
        perms = []
        if permute:
            for perm in permutations(ab):
                perm_seqs = ["".join([perm[x-1] for x in seq]) for seq in seqs]
                identities = self.align(perm_seqs)
                identities = identities * len_score
                scores.append(identities)
                perms.append(perm)
        else:
            seqs = ["".join([ab[x-1] for x in seq]) for seq in seqs]
            identities = self.align(seqs)
            identities = identities * len_score
            scores.append(identities)
            perms.append([x for x in ab])
        return scores, perms
        
        
    def align(self,
              sequences: List[str]
              ) -> ndarray:
        """
        Get the align score given a batch of sequences.

        Parameters
        ----------
        sequences : List[str]
            A list of the sequences.
        Returns
        -------
        ndarray
            A array with same length of the sequences, gives the identity
            scores.

        """
        identities = []
        for seq in sequences:
            hits = self.align_seq(seq)
            identity = self._match_score(hits[0])/len(seq) if len(hits)>0 else 0  #Only use the first hit.
            identities.append(identity)
        return np.array(identities)
    
    def _match_score(self, alignment: Optional[bwapy.libbwa.Alignment]) -> int:
        """
        Get the match score given a cigar string.
    
        Parameters
        ----------
        alignment : Optional[Alignment]
            A alignment instance from the bwapy mapping result.
    
        Returns
        -------
        int
            The nubmer of matched bases.
    
        """
        cigar_s = alignment.cigar
        match_hits = re.findall(r'(\d+)M', cigar_s)
        n_match = sum([int(x) for x in match_hits])
        return n_match - alignment.NM

if __name__ == "__main__":
    reference_f = '/home/heavens/twilight_hdd1/m6A_Nanopore/references.fasta'
    test_seqs = ['AAGAUUGUGUUGGUCAGUACUCGGUCGACGGGGUGAUUCCUUCAGCCCAGCCUGAUGAAUGCCGACACCUACCGGCUUUUAAGACGCCCAGGUACCAACGAUCAUCGUCGACCGUAAGU',
                'AUAGAUUGUGUUUGUUAGUCGCUUGGUACGCCGAGCUGAAUGAUUGUGGAUAUCGCGAAGCAGGUUGAGCGCAACCCUGGACGACCGUAUUUCGGCGUCAAUAAGCGUCGUGAAGGCAGGUGAAUAUGAGCAUGGCGUUGGGGCUUUCCGAUAUCAAAGGCGCAGAUGAGCGGUCUGGUAUACCGUGGUUUGUGACUGGCAACGGCUUUAACGUUGGCGAUACGCUGCGUUAUCUGCAGGCAAUUGAGCGGCUGGAGAAAAACUGGCUGUCGAUCCGUCCGAACCGGGCGCGAAAUGCCGAAAGUCGAGCGUCCAGCAGGCGUGGCAUAAUGGAUAAUAAAACUGCCCAGCGCGUCGCGAAGAUGAUGAUGUUAAAGAGAUCCGUUUGGAUGAUUGAAGAACUGCGUGUCAGUAUUCGCCCAACGGUUACGGUGCCUUAUCCUAUCUCCGAUAAGCGUAUUUUACAGGCCAUGGAUCAGAUUACGGCCUAAAGCCAGGCGACGAUGAGCUAAAGAGAUCCAGUACAAGUUCGGAACAAAAAGGAAAUUAUGAAUACAACGCCAUUCCAGCGCUUUAAUGAAACCAUUGCUGGCGGUGAAUACACCCAAUGACAGUGGCUGGCGCAGCGUGAUUUUUCCGGCGAACAACCAUCAUCGGAUUCGCGACCGGCGGGCAAUGCGGUGAUCCCACAUCGACGCGGCCUGUCGCAUCGCCAAGCGCAGGCGUUCCACUGAUUGCGGCGGUAUAAUGGGUCACGCCUUCGACGCCGUUUUUGCGCCGUUGAUUGCCCGGCAUCCCGCGCUACCACACUAUCUGCCAACCGGACGCGCUGAAGCCAUCGAUCUCGCGGAUCGCGAACCAGUCUGGCAUAUUCCGGCUGAGAAAUCUGGCUGGAAGGUCGGUCAACUAUUGCGGCGAAAAUGCCCGUCACCUGCGUCAAUCCGCCAGGCGAAAGAAAACAUUAACGGCUAUCGUUGCGAAGACCCCACCAUGCAGCAGCGCACUAUCGCGUUAUUCCGGCGCGUGACAAACGACGACACCGAUGCGUGCUGGCUGAGUUUUUCCUGGUUUGUUCCAGUACUGCGCCACCUUAACGGCGAUACACGUUUUGCCUGUGUUGAAGAGGGCAUGGGACGGUGGAGCGUUAUCUGUACCGAGAUUGCCGGCGAGCUGCCUGCGUGAUGAAACAGGGUAAGCACCGCGCGGUAAAGAUUUUAUUUCACGUUGGAUAUCCCCACACGCGAUUAUUAGAAACGCAUUGGCGGUAUUAGCAGGCAGAUACGACGCUGCGCAGUGCAAUUAGAACAGAGAGCGUUCGCGAAAAAAUCCCGUCUGCGGGUGAACGGGCAAGUGCGACCACCAUAUCUCAUAAAUUUAAUGAAUAUUUAUAAAAGCAAAAUCGUUGUCGCUCACGGUUUCAUUCAGGACGCGCUAUGGGCGGCAAGUAUUCCGGCCUGCAAAUUGGUAUUUUACUGGUUAAGU',
                'AGACUUGUGUUUGGCGGUCGCUUCUAAACCAGCGUAAAAGA',
                'CAGAUAGAUAGUGUUUGUUAGUCGCUUUGGUGCGAAACUCCACUUAUCUACUUACUACGGUCAAAUUAACGAUCAAGCUUUAAGUACGUCGCCAGAUUGUCUGGUUUAACCCUCAAGUGCGGGACAAUUAUUGGGACGAAUACUACGAGUACGGUCGUGUGGCCAUCUCCACCGUCCCCCGCUGUAAUGCUUUUUUUACUCGUCUUAAGGAUAGGUAUCAUCUGACUAGUACCGGCGUGCAAAAGCUUUUUCCAGGAAAGUAGUCAGGAUGAGGCUGGUGAACUGUUACGAAAGAUGAAGGUAAGACCUCAUCCUACAUUUCACAGGCAAGCCAUUAAAGUGCACUGUCUCGUGGCGAUUGCAACUUCGAUCGGCAUCUGAUCAUUGGCCAGCCCUUAGCCGUGUUUCGGUGUGAAGCAGCGUGGCGGGAAUCGCCAUCAUGAUCCCCGCAUGUAGGGCCCUAUCGGAAUUACACCGUCCUAAUGGUUUUUGAAUUGCGGUGGCGCCAGAAAAGA',
                'AGAUUGUGUUUGUUAGUUGCUUAACAAAAUGGUACAGAGAUGAUAGAAAGGCCUGGAGCCGUCUUCCGGGAAUCCAGAACAGAAUCAUCAACAUUUCCACGAAUCUCAAAGAGCUUAUUGAUUUUGGGCUGAGGCAGCUUGGAAGAAAGCAAUCUCCCAGGCAAAAGCAGCUGUUUUUUUAUCCGUCGAACGGUACUGCACUGCUGUCAUGGGCCGACAGGUUCCAGUAGCUCCUUACCAUGCGAACCGGAUCUACCAGUUUGCCAUCUACUCUGACAUUUUGAAGCUCAGGUGCACCAUUUAUUACACUAACUGUGCAGAUUACCAAUCCAAACUGCUUUCAAUUCGGACACAAAAAAAAGGUUUUACAGGUGCCAGCGAGGAUAAGGCUGGUUUAGUCGAACAAGCCAAUGGAGCAUACACUCUUAUGGACGGCUCCGCCUCCCAAAGGCAAGAAAUGUUGUCUUGCUACUGACAGCGGUUCACCACAACGGCUUGGAGAAUCCUGAACAUAAACGAACAUCAGACGUAUUGUUUAUUUGUGCCACUACAGAAAACCUAGGUCUGCAACUGAAAACAUUUUUGCAGAAUCCCAAUGACCAUUUUAACCAUGUUGUAGAACGAUCGCUUAAAGAAAAGGGUGGAUCUGACGACGUUUUAUUUAGGAAAGAAGCAGAACGAAUUAAGAAAAAUCUGAGUGUGCACAUAGAUGUCUAUAAUGCGCUUAUUCAUUCUGCGAAAUUUGAAAUGUGGGACAGCUGAAAUCAAAUGUCCAGUUAGUUUGUGCACACAGGGUUUUUGUUCCAACCUUGACGGACCGAGGUUAAUGAAAUGACUGUUCGGGAUCUGCCGGAUGAAAUCAAGCAAGAAUGGAUGUCCAGCAGCAAAAUACUGCAAAGGAGCAAAGCCAUUUGAAUGUCAUAAUUUUUUCGAUCAUCUCCGAUUAAGCGGAAGAUGAAACAACCGAAUAUUGACGAGGACCCUUUUCAUUUAACUUGUAUCAUAUUGCGAUCGAAGAAAAGGAG',
                'AGAUUGUGUUUGUUAGUUGCUUAACAAAAUGGUACAGAGAUGAUAGAAAGGCCUGGAGCCGUCUUCCGGGAAUCCAGAACAGAAUCAUCAACAUUUCCACGAAUCUCAAAGAGCUUAUUGAUUUUGGGCUGAGGCAGCUUGGAAGAAAGCAAUCUCCCAGGCAAAAGCAGCUGUUUUUUUAUCCGUCGAACGGUACUGCACUGCUGUCAUGGGCCGACAGGUUCCAGUAGCUCCUUACCAUGCGAACCGGAUCUACCAGUUUGCCAUCUACUCUGACAUUUUGAAGCUCAGGUGCACCAUUUAUUACACUAACUGUGCAGAUUACCAAUCCAAACUGCUUUCAAUUCGGACACAAAAAAAAGGUUUUACAGGUGCCAGCGAGGAUAAGGCUGGUUUAGUCGAACAAGCCAAUGGAGCAUACACUCUUAUGGACGGCUCCGCCUCCCAAAGGCAAGAAAUGUUGUCUUGCUACUGACAGCGGUUCACCACAACGGCUUGGAGAAUCCUGAACAUAAACGAACAUCAGACGUAUUGUUUAUUUGUGCCACUACAGAAAACCUAGGUCUGCAACUGAAAACAUUUUUGCAGAAUCCCAAUGACCAUUUUAACCAUGUUGUAGAACGAUCGCUUAAAGAAAAGGGUGGAUCUGACGACGUUUUAUUUAGGAAAGAAGCAGAACGAAUUAAGAAAAAUCUGAGUGUGCACAUAGAUGUCUAUAAUGCGCUUAUUCAUUCUGCGAAAUUUGAAAUGUGGGACAGCUGAAAUCAAAUGUCCAGUUAGUUUGUGCACACAGGGUUUUUGUUCCAACCUUGACGGACCGAGGUUAAUGAAAUGACUGUUCGGGAUCUGCCGGAUGAAAUCAAGCAAGAAUGGAUGUCCAGCAGCAAAAUACUGCAAAGGAGCAAAGCCAUUUGAAUGUCAUAAUUUUUUCGAUCAUCUCCGAUUAAGCGGAAGAUGAAACAACCGAAUAUUGACGAGGACCCUUUUCAUUUAACUUGUAUCAUAUUGCGAUCGAAGAAAAGGAG']
    
    aln = MetricAligner(reference_f)
    scores = aln.align(test_seqs)
    print(scores)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 11 19:55:38 2022

@author: haotian
"""

import os
import sys
import toml
import numpy as np
import argparse

def merge(args):
    fs = args.fs
    keys = args.keys
    collections = {x+'.npy':[] for x in keys}
    for f_i,sub_f in enumerate(fs):
        print("Collecting from %s..."%(sub_f))
        npy_files = [x for x in os.listdir(sub_f) if x.endswith(".npy")]
        for key in collections.keys():
            if key not in npy_files:
                print("The chunk %s is not in %s, stop collecting it."%(key,sub_f))
                del collections[key]
                continue
            else:
                print("Collecting %s..."%(key))
                pieces = np.load(os.path.join(sub_f,key),mmap_mode = "r")
                print("Collected %d instances"%(len(pieces)))
                if args.max is not None and len(pieces) > args.max[f_i]:
                    print("Thresholding it to %d instances"%(args.max[f_i]))
                    pieces = pieces[:args.max[f_i]]
                collections[key].append(pieces)
    shapes = [sum([len(y) for y in x]) for x in collections.values()]
    assert len(np.unique(shapes)) == 1

    for key in collections.keys():
        print("merge %s"%(key))
        if collections[key][0].ndim > 1:
            np.save(os.path.join(args.output,key),np.vstack(tuple(collections[key])))
        else:
            np.save(os.path.join(args.output,key),np.hstack(tuple(collections[key])))

def main(args):
    merge(args)
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training RHMM model')
    parser.add_argument("-i","--input", type = str, required = True,
                        help = "Data folder contains npy file, separated by comma.")
    parser.add_argument("-o","--output",required = True, type = str,
                        help = "The output folder to store the merged dataset.")
    parser.add_argument("-k","--key", type = str, default = "chunks,path,seqs,seq_lens,durations",
                        help = "The name of npy items need to be collected, separated by comma.")
    parser.add_argument("-m","--max", default = None,
                        help = "The maximum number of instances to be include in each dataset, can be a list of int separated by commas specify maximum number for each datasets.")
    args = parser.parse_args(sys.argv[1:])
    os.makedirs(args.output,exist_ok = True)
    args.fs = args.input.strip().split(',')
    args.keys = args.key.strip().split(',')
    if args.max:
        if ',' in args.max:
            args.max = [int(x) for x in args.max.split(',')]
        else:
            args.max = [int(args.max)] * len(args.fs)
        assert len(args.fs) == len(args.max)
    else:
        args.max = None
    main(args)



"""
Created on Thu Mar  4 19:33:04 2021

@author: Haotian Teng
"""
import numpy as np
import torchvision
import torch
from torch import nn
import torch.utils.data as data
from torchvision import transforms
from matplotlib import pyplot as plt
from typing import Dict,Callable

class Dataset(data.Dataset):
    """
    Nanopore DNA/RNA chunks dataset
    """
    def __init__(self, 
                 chunks:np.ndarray, 
                 seq:np.ndarray = None, 
                 seq_len:np.ndarray = None, 
                 chunks_len:np.ndarray = None,
                 transform:torchvision.transforms.transforms.Compose=None,
                 seq_padding:bool = True):
        """
        Generate a training dataset 

        Parameters
        ----------
        chunks : np.ndarray
            A N-by-M matrix, where N is the number of samples and M is the
            length of chunks.
        seq : np.ndarray, optional
            A N-by-D matrix, where N is the number of samples and D is the
            length of sequence. If this is given a seq_len vector must also
            be given to specify the length of the sequence. The default is None.
        seq_len : np.ndarray, optional
            A length N vector, gives the actual length of the sequence. 
            If not given then no label is used. The default is None.
        chunks_len: np.ndarray
            A length N vector, where N is the number of samples, indicate the
            length of the signal. If [M]*N vector will be used.
        transform : torchvision.transforms.transforms.Compose, optional
            DESCRIPTION. The default is None.
        seq_padding: bool, optional
            Default is True, if padding the sequence.

        Returns
        -------
        None.

        """        
        if seq_padding & (seq is not None):
            l_max = max(seq_len)
            seq = np.array([x+'$'*(l_max-len(x)) for x in seq])
        self.chunks = chunks
        if chunks_len is None:
            self.chunks_len = np.array([[chunks.shape[1]]]*chunks.shape[0],dtype = np.int64)
        else:
            self.chunks_len = chunks_len[:,None].astype(np.int64)
        self.seq = seq
        if seq_len is not None:
            self.seq_len = seq_len[:,None].astype(np.int64)
        assert(not (seq is None)^(seq_len is None) )
        self.transform = transform
                
    def __len__(self):
        return len(self.chunks)

    def __getitem__(self, idx):
        if self.seq is None:
            sample = {'signal': self.chunks[idx][None,:].astype(np.float32),
                      'signal_len': self.chunks_len[idx]}
        else:
            sample = {'signal': self.chunks[idx][None,:].astype(np.float32), 
                      'signal_len': self.chunks_len[idx],
                      'seq': self.seq[idx],
                      'seq_len': self.seq_len[idx]}

        if self.transform:
            sample = self.transform(sample)
        return sample

class NumIndex(object):
    """
    Convert string sequence to integer index array
    
    Parameters
    ----------
    alphabet_dict : Dict
        A dictionary map the string to index array.
            
    """
    def __init__(self,alphabet_dict:Dict):
        self.alphabet_dict = alphabet_dict
        self.alphabet_dict['$'] = 0 #Default padding character
        self.alphabet_dict['N'] = 0 #Remove unknonw nucleotide
    def __call__(self,sample:Dict):
        seq = list(sample['seq'])
        seq = [self.alphabet_dict[x] for x in seq]
        seq = np.array(seq,dtype = np.long)
        return {key:value if key != 'seq' else seq for key, value in sample.items()}

class ToTensor(object):
    def __call__(self,sample:Dict):
        func = torch.from_numpy
        return {key:func(value) for key, value in sample.items()}

def show_sample(sample:Dict, idx = 0):
    """
    Plot a sample signal, if input is a batch, plot the idx-th one.

    Parameters
    ----------
    sample : Dict
        A dictionary contain the samples.

    Returns
    -------
    None.

    """
    signal = sample['signal']
    if len(signal.shape) > 1:
        plt.plot(signal[idx,:])
    else:
        plt.plot(signal)

if __name__ == "__main__":
    print("Load dataset.")
    folder = "/home/heavens/bridge_scratch/ELIGOS_dataset/IVT/control/kmers_guppy_4000_noise/"
    chunks = np.load(folder + "chunks.npy")
    reference = np.load(folder + "seqs_re.npy")
    ref_len = np.load(folder + "seq_re_lens.npy")
    plt.hist(ref_len[ref_len<chunks.shape[1]],bins = 200)
    alphabet_dict = {'A':1,'C':2,'G':3,'T':4,'M':5}
    print("Filt dataset.")
    # chunks,reference,ref_len = filt(RNA_FILTER_CONFIG,chunks,reference,ref_len)
    print("Create dataset.")
    dataset = Dataset(chunks,
                      seq = reference,
                      seq_len = ref_len,
                      transform = transforms.Compose([NumIndex(alphabet_dict),ToTensor()]))
    loader = data.DataLoader(dataset,batch_size = 200,shuffle = True, num_workers = 4)
    for i_batch, sample_batch in enumerate(loader):
        if i_batch == 10:
            show_sample(sample_batch)
            break
    



"""
Created on Tue Dec 28 13:33:50 2021

@author: Haotian Teng
"""
import torch

class NearstEmbedding(torch.autograd.Function):
    """Get the nearest embedding of given input, used in VQ-VAE:
        https://arxiv.org/pdf/1711.00937.pdf
    """
    @staticmethod
    def forward(ctx,input,embd_weight):
        """
        Find the nearest neighbour of the embedding variable.

        Parameters
        ----------
        ctx : Object.
            A object to stash information for backward computation, save 
            arbitrary objects for use in backward computation by ctx.cave_for_backward method.
        input : torch.tensor
            The input tensor with size (*,embedding_dim)
        embd_weight : torch.tensor
            The embedding weight tensor with size (num_embeddings,embedding_dim)

        Returns
        -------
        The nearest embedding vector.

        """
        index = vq_idx(input,embd_weight)
        ctx.mark_non_differentiable(index)
        return torch.index_select(embd_weight,dim = 0,index = index).view_as(input)
    
    @staticmethod
    def backward(ctx,grad_post):
        """
        Propgate the gradient from posterior layer back to the previous layer,
        dL/dq = dL/de.
        """
        grad_input, grad_embd = None,None
        if ctx.needs_input_grad[0]:
            grad_input = grad_post.clone()
        return grad_input,grad_embd

class NearstEmbeddingIndex(torch.autograd.Function):
    """Get the indexs of the nearest embedding of given input, no gradient will
    be propagate back.
    """
    @staticmethod
    def forward(ctx,input,embd_weight):
        """
        Find the nearest neighbour of the embedding variable.

        Parameters
        ----------
        ctx : Object.
            A object to stash information for backward computation, save 
            arbitrary objects for use in backward computation by ctx.cave_for_backward method.
        input : torch.tensor
            The input tensor with size (*,embedding_dim)
        weight : torch.tensor
            The embedding weight tensor with size (num_embeddings,embedding_dim)

        Returns
        -------
        The nearest embedding vector.

        """
        with torch.no_grad():
            if input.shape[-1] != embd_weight.shape[-1]:
                raise ValueError("Input tensor shape %d is not consistent with the embedding %d."%(input.shape[-1],embd_weight.shape[-1]))
            distance = torch.cdist(input.flatten(end_dim = -2),embd_weight)
            index = torch.argmin(distance,dim = 1)
            ctx.mark_non_differentiable(index)
            return index
    
    @staticmethod
    def backward(ctx,grad_post):
        """
        No gradient since this is for training embedding.
        """
        raise RuntimeError('Trying to call `.grad()` on graph containing '
            '`NearstEmbeddingIndex`. The function `NearstEmbeddingIndex` '
            'is not differentiable. Use `NearstEmbedding` '
            'if you want a straight-through estimator of the gradient.')

def vq_idx(x,embd_weight):    
    return NearstEmbeddingIndex().apply(x,embd_weight)

def vq(x:torch.Tensor, embd_weight:torch.Tensor):
    """
    Apply vector quantised to the given input x and a coding book embed_weight

    Parameters
    ----------
    x : torch.Tensor
        The input tensor with shape [*,C], where C is the size of embedding.
    embd_weight : torch.Tensor
        The coding book of the embedding with shape [N,C] where N is the number
        of embeddings and C is the size of embedding.

    Returns
    -------
    e : torch.Tensor
        The quantised embedding with same shape as input x.
    e_shadow : torch.Tensor
        The shadow tensor of e with same shape as input x.

    """
    idx = NearstEmbeddingIndex().apply(x,embd_weight.detach())
    e_shadow = torch.index_select(embd_weight,dim = 0,index = idx).view_as(x)
    e = NearstEmbedding().apply(x, embd_weight.detach())
    return e,e_shadow

def loss(x,embd_weight):
    e,e_shadow = vq(x,embd_weight)
    mse_loss = torch.nn.MSELoss(reduction = "mean")
    sg_q = x.detach()
    sg_e = e.detach()
    rc_signal = e
    rc_loss = mse_loss(rc_signal,x)
    embedding_loss = mse_loss(sg_q,e_shadow)
    commitment_loss = mse_loss(sg_e,x)
    return rc_loss, embedding_loss, commitment_loss

if __name__ == "__main__":
    import torchviz
    from matplotlib import pyplot as plt
    from matplotlib import image
    print("Run testing code for Vector-quantization.")
    N_EMBD = 10
    EMBD_DIM = 3
    torch.manual_seed(1992)
    embd = torch.nn.Embedding(N_EMBD,EMBD_DIM)
    embd.weight.data.uniform_(-1./N_EMBD, 1./N_EMBD)
    embd_weight = embd.weight
    x = torch.rand((2,2,EMBD_DIM),requires_grad=True, dtype = torch.float)
    
    idx = NearstEmbeddingIndex.apply(x,embd_weight)
    rc_loss, embedding_loss, commitment_loss= loss(x,embd_weight)
    grad_x,grad_embd = torch.autograd.grad((rc_loss, embedding_loss, commitment_loss), (x,embd_weight), create_graph=True)
    torchviz.make_dot((grad_x,grad_embd, x, embd_weight, rc_loss, embedding_loss, commitment_loss), 
                      params={"grad_x": grad_x, 
                              "grad_embd":grad_embd, 
                              "x": x, 
                              "embedding_weight":embd_weight, 
                              "rc_loss": rc_loss, 
                              "embedding_loss":embedding_loss, 
                              "commitment_loss":commitment_loss}).render("./all_loss", format="png")
    
    e = NearstEmbedding().apply(x,embd_weight)
    nn_grad ,_ = torch.autograd.grad((e.sum()), (x,embd_weight), create_graph=True,allow_unused=True )
    torchviz.make_dot((nn_grad, x, embd_weight,e), params={"grad_x": nn_grad, "x": x, "embedding_weight":embd_weight,"e":e}).render("./nearestembd", format="png")
    img = image.imread("./nearestembd.png")
    imgplot = plt.imshow(img)


"""
Created on Fri Nov 19 22:51:28 2021

@author: Haotian Teng
"""
import numpy as np
from itertools import product,repeat

dataset = "/home/heavens/bridge_scratch/m6A_Nanopore/merged_dataset_diff/"
chunks_f = dataset + "chunks.npy"
seq_lens_f = dataset + "seq_lens.npy"
seqs_f = dataset + "seqs.npy"
chunks = np.load(chunks_f)
seq_lens = np.load(seq_lens_f)
seqs = np.load(seqs_f)

def check_ratio(sequences):
    control_count = 0
    m_count = 0
    other_count = 0
    for seq in sequences:
        if 'A' in seq:
            if 'M' in seq:
                raise("ValueError: The read has both M and A in sequence.")
            control_count +=1
        elif 'M' in seq:
            m_count += 1
        else:
            other_count +=1
    return control_count,m_count, control_count + m_count+other_count

def check_Mkmer(sequences,k = 5):
    N = ['M','C','G','T']
    kmers = product(*repeat(N,k))
    kmer_dict = {}
    for kmer in kmers:
        kmer_dict[''.join(kmer)] = 0
    for seq in sequences:
        occurs = [i for i,c in enumerate(seq) if c=='M']
        for o in occurs:
            curr_kmer = seq[o:o+k]
            if len(curr_kmer) == k:
                kmer_dict[curr_kmer] +=1
    return kmer_dict

c,m,all = check_ratio(seqs)
kmer_dict = check_Mkmer(seqs)


"""
Created on Thu Mar 11 16:07:12 2021

@author: Haotian Teng
"""
import os
import toml
import torch
import itertools
import numpy as np
from typing import Union,Dict
from torch.utils.data.dataloader import DataLoader
from xron.xron_model import CRNN, REVCNN, CONFIG,DECODER_CONFIG,MM


class Trainer(object):
    def __init__(self,
                 train_dataloader:DataLoader,
                 nets:Dict[str,Union[CRNN,REVCNN,MM]],
                 config:Union[CONFIG,DECODER_CONFIG],
                 device:str = None,
                 eval_dataloader:DataLoader = None):
        """

        Parameters
        ----------
        train_dataloader : DataLoader
            Training dataloader.
        nets : Dict[str,Union[CRNN,REVCNN]]
            A CRNN or REVCNN network instance.
        device: str
            The device used to train the model, can be 'cpu' or 'cuda'.
            Default is None, use cuda device if it's available.
        config: CONFIG
            A CONFIG class contains training configurations. Need to contain
            at least these parameters: keep_record, device and grad_norm.
        eval_dataloader : DataLoader, optional
            Evaluation dataloader, if None training dataloader will be used.
            The default is None.

        Returns
        -------
        None.

        """
        self.train_ds = train_dataloader
        self.device = self._get_device(device)
        if eval_dataloader is None:
            self.eval_ds = self.train_ds
        else:
            self.eval_ds = eval_dataloader
        self.nets = nets
        paras = [x.parameters() for x in self.nets.values()]
        self.parameters = itertools.chain(*paras)
        for net in self.nets.values():
            net.to(self.device)
        self.global_step = 0
        self.save_list = []
        self.keep_record = config.TRAIN['keep_record']
        self.grad_norm = config.TRAIN['grad_norm']
        self.config = config
        self.losses = []
        self.errors = []
    
    def reload_data(self,train_dataloader, eval_dataloader = None):
        self.train_ds = train_dataloader
        if eval_dataloader is None:
            self.eval_ds = train_dataloader
        else:
            self.eval_ds = eval_dataloader
    
    def _get_device(self,device):
        if device is None:
            if torch.cuda.is_available():
                return torch.device('cuda')
            else:
                return torch.device('cpu')
        else:
            return torch.device(device)

    def _update_records(self):
        record_file = os.path.join(self.save_folder,'records.toml')
        with open(record_file,'w+') as f:
            toml.dump(self.records,f)

    def save(self):
        ckpt_file = os.path.join(self.save_folder,'checkpoint')
        current_ckpt = 'ckpt-'+str(self.global_step)
        model_file = os.path.join(self.save_folder,current_ckpt)
        self.save_list.append(current_ckpt)
        if not os.path.isdir(self.save_folder):
            os.mkdir(self.save_folder)
        if len(self.save_list) > self.keep_record:
            os.remove(os.path.join(self.save_folder,self.save_list[0]))
            self.save_list = self.save_list[1:]
        if os.path.isfile(model_file):
            os.remove(model_file)
        with open(ckpt_file,'w+') as f:
            f.write("latest checkpoint:" + current_ckpt + '\n')
            for path in self.save_list:
                f.write("checkpoint file:" + path + '\n')
        net_dict = {key:net.state_dict() for key,net in self.nets.items()}
        torch.save(net_dict,model_file)
    
    def save_loss(self):
        loss_file = os.path.join(self.save_folder,'losses.csv')
        error_file = os.path.join(self.save_folder,'errors.csv')
        if len(self.losses):
            with open(loss_file,'a+') as f:
                f.write('\n'.join([str(x) for x in self.losses]))
                f.write('\n')
        if len(self.errors):
            with open(error_file,'a+') as f:
                f.write('\n'.join([str(x) for x in self.errors]))
                f.write('\n')
        self.losses = []
        self.errors = []
    
    def _save_config(self):
        config_file = os.path.join(self.save_folder,'config.toml')
        config_modules = [x for x in self.config.__dir__() if not x .startswith('_')][::-1]
        config_dict = {x:getattr(self.config,x) for x in config_modules}
        with open(config_file,'w+') as f:
            toml.dump(config_dict,f)
    
    def load(self,save_folder,update_global_step = True):
        self.save_folder = save_folder
        ckpt_file = os.path.join(save_folder,'checkpoint')
        with open(ckpt_file,'r') as f:
            latest_ckpt = f.readline().strip().split(':')[1]
            if update_global_step:
                self.global_step = int(latest_ckpt.split('-')[1])
        ckpt = torch.load(os.path.join(save_folder,latest_ckpt),map_location=self.device)
        for key,net in ckpt.items():
            if key in self.nets.keys():
                try:
                    self.nets[key].load_state_dict(net,strict = True)
                except RuntimeError:
                    print(f"Exact loading {key} failed, try load loosely.")
                    self.nets[key].load_state_dict(net,strict = False)
                self.nets[key].to(self.device)
            else:
                print("%s net is defined in the checkpoint but is not imported because it's not defined in the model."%(key))

class DeviceDataLoader():
    """Wrap a dataloader to move data to a device"""
    def __init__(self, dataloader, device = None):
        self.dataloader = dataloader
        if device is None:
            device = self.get_default_device()
        else:
            device = torch.device(device)
        self.device = device
    
    def __iter__(self):
        """Yield a batch of data after moving it to device"""
        for b in self.dataloader:
            yield self._to_device(b, self.device)
    
    def __len__(self):
        """Number of batches"""
        return len(self.dataloader)
    
    def _to_device(self,data,device):
        if isinstance(data, (list,tuple)):
            return [self._to_device(x,device) for x in data]
        if isinstance(data, (dict)):
            temp_dict = {}
            for key in data.keys():
                temp_dict[key] = self._to_device(data[key],device)
            return temp_dict
        return data.to(device, non_blocking=True)
    
    def get_default_device(self):
        if torch.cuda.is_available():
            return torch.device('cuda')
        else:
            return torch.device('cpu')

def load_config(config_file):
    class CONFIG(object):
        pass
    with open(config_file,'r') as f:
        config_dict = toml.load(f)
    config = CONFIG()
    for k,v in config_dict.items():
        setattr(config,k,v)
    return config


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  9 07:09:45 2022

@author: heavens
"""
from faulthandler import dump_traceback_later
import os
import sys
import toml
import random
import argparse
import itertools
from typing import List
from xron.utils.seq_op import Methylation_DP_Aligner
from xron.nrhmm.kmer2seq import fixing_looping_path
from functools import partial
from tqdm import tqdm
import numpy as np


class LinkingGraph(object):
    def __init__(self,config):
        self.config = config
        self.idx2kmer = config['idx2kmer']
        self.kmer2idx = config['kmer2idx_dict']
        self.exploration = config['exploration']
        self.n_kmer = len(self.idx2kmer)
        self.nodes = {}
        self.invariant_kmers = []
        self.aligner = Methylation_DP_Aligner(base_alternation=config['base_alternation'],)
        
    def build_invariant_kmers(self, variant_bases):
        for i in np.arange(self.n_kmer):
            if not any([x in self.idx2kmer[i] for x in variant_bases]):
                self.invariant_kmers.append(i)
    
    def add_reads(self,segments,decoded,durations,min_seg_len = 500,transition = None):
        if len(self.invariant_kmers) == 0:
            raise ValueError("Call build_invariant_kmers first.")
        if transition is not None:
            decoded = np.asarray([[transition(x) for x in d] for d in decoded])
        mask = np.isin(decoded,self.invariant_kmers)
        breakings = np.where(mask) #Tuple (x,y) x is a Length N batch idx vector, y is a length N vector index of kmer in each segment. N is the number of invariant kmers
        start,end,duration = self.boundering(breakings)
        #Start is a 2XN start idxes [batch_idx, sig_idx] of a invariant kmer
        #End is a 2XN vector the end idx [batch_idx, sig_idx] of a invariant kmer
        curr = 1
        N = len(start[0])
        prev = 0
        while curr < N:
            start_index = start[:,prev]
            s_i,s_j = start_index
            next_i = start[0][curr]
            if next_i != s_i:
                #Skip because we are not at different batch
                prev = curr
                curr += 1
                continue
            if (start[1][curr] - s_j) < min_seg_len:
                curr += 1
                continue
            if (end[1][curr] > durations[s_i]):
                #Skip because this segment at the padding region
                prev = curr + 1
                curr += 2
                continue
            curr_kmer = decoded[s_i,s_j]
            curr_segment = segments[s_i,s_j:end[1][curr]+1]
            curr_path = decoded[s_i,s_j:end[1][curr]+1]
            if curr_kmer not in self.nodes.keys():
                node = KmerNode(curr_kmer)
                self.nodes[curr_kmer] = node
            self.nodes[curr_kmer].add_segment(curr_segment,curr_path)
            prev = curr
            curr += 1
    
    def boundering(self,breakings):
        x,y = breakings
        start_mask = np.logical_or((x[1:]-x[:-1]) != 0,(y[1:]-y[:-1]) != 1)
        start = np.asarray([np.insert(x[1:][start_mask],0,x[0]),np.insert(y[1:][start_mask],0,y[0])])
        end_mask = np.logical_or((x[:-1]-x[1:]) != 0,(y[:-1]-y[1:]) != -1)
        end = np.asarray([np.append(x[:-1][end_mask],x[-1]),np.append(y[:-1][end_mask],y[-1])])
        duration = [end[0] - start[0],end[1]-start[1]+1]
        assert np.all(duration[0] == 0)
        return start, end, duration[1]

    def sampling_once(self,max_len,add_noise = 0):
        segment = np.asarray([])
        path = np.asarray([],dtype = np.int16)
        path_idx = np.asarray([],dtype = np.int16)
        curr_i = 0
        p = np.asarray([x.weight for x in self.nodes.values()])
        if np.random.random() < exploration:
            kmer = np.random.choice(list(self.nodes.keys()))#random choose a kmer to start
        else:
            kmer = list(self.nodes.keys())[np.argmax(p)]
        while len(segment)<max_len:
            if kmer not in self.nodes.keys():
                break #Sampling to a dead end
            node = self.nodes[kmer]
            curr_seg,curr_path,kmer = node.sampling(exploration = self.exploration)
            if len(curr_seg) + len(segment) > max_len:
                break #The maximum length has been reached
            segment = np.concatenate((segment,curr_seg),axis = 0)
            path = np.concatenate((path,curr_path),axis = 0)
            path_idx = np.concatenate((path_idx,np.full(len(curr_path),curr_i)),axis = 0)
            curr_i += 1
        if add_noise > 0:
            segment += np.random.normal(loc = 0.0,scale = add_noise,size = len(segment))
        if len(segment)<max_len:
            segment = np.concatenate((segment,np.full(max_len-len(segment),0)),axis = 0)
            path = np.concatenate((path,np.full(max_len-len(path),-1)),axis = 0)
            path_idx = np.concatenate((path_idx,np.full(max_len-len(path_idx),-1)),axis = 0)
        return segment[:max_len],path[:max_len],path_idx[:max_len]
    
    def sampling(self,max_len,n_times):
        segments = []
        paths = []
        path_idxs = []
        for i in tqdm(range(n_times),desc = "Sampling"):
            seg,path,path_idx = self.sampling_once(max_len)
            segments.append(seg)
            paths.append(path)
            path_idxs.append(path_idx)
        return np.stack(segments,axis = 0),np.stack(paths,axis = 0),np.stack(path_idxs,axis = 0)
        
class KmerNode(object):
    def __init__(self,kmer):
        self.kmer = kmer
        self.segments = []
        self.visits = [] #Sampling times
        self.weight = 0.
        self.epsilon = 1e-6
        self.head_stickers = []
        self.tail_stickers = []
        self.paths = []
        self.next_kmers = []
        self.next_nodes = []
        
    def add_segment(self,segment,path):
        """
        Add segement to the node, both the start and end kmer need to be
        included

        Parameters
        ----------
        segment : np.array
            The segment of the signal, the start and the end invariant kmer
            need to be included.
        path : np.array
            The corresponding kmer path.
        """
        assert len(segment) == len(path)
        self.segments.append(segment[path != path[-1]])
        self.head_stickers.append(segment[path == path[0]])
        self.tail_stickers.append(segment[path == path[-1]])
        self.paths.append(path[path != path[-1]])
        self.next_kmers.append(path[-1])
        self.visits.append(self.epsilon)
        self.weight += 1
        
    def sampling(self,exploration = 0.1):
        p = np.sqrt(1/np.asarray(self.visits)) #UCB score.
        p = p/sum(p)
        if np.random.random() < exploration:
            i = np.random.choice(len(self.segments))
        else:
            i = np.argmax(p)
        self.visits[i] += 1
        self.weight = len(self.visits)/np.sqrt(sum(self.visits))
        return self.segments[i],self.paths[i],self.next_kmers[i]
    
def kmers2seq(kmers,idx2kmer):
    merged = [g for g,_ in itertools.groupby(kmers)]
    seqs = [idx2kmer[x][0] for x in merged]
    return ''.join(seqs) + idx2kmer[merged[-1]][1:]

def load_data(data_f,mmap_mode = "r",max_n = None,shuffle = False):
    chunks = np.load(os.path.join(data_f,"chunks.npy"),mmap_mode = mmap_mode)
    paths = np.load(os.path.join(data_f,"path.npy"),mmap_mode = mmap_mode)
    delooped = False    
    if os.path.exists(os.path.join(data_f,"path_fix.npy")):
        paths_fix = np.load(os.path.join(data_f,"path_fix.npy"),mmap_mode = mmap_mode)
        if len(paths_fix) == len(paths):
            print("Loopping fixed path has been found, using it")   
            paths = paths_fix
            delooped = True
        else:
            delooped = False
    seqs = np.load(os.path.join(data_f,"seqs.npy"),mmap_mode = mmap_mode)
    seq_lens = np.load(os.path.join(data_f,"seq_lens.npy"),mmap_mode = mmap_mode)
    durations = np.load(os.path.join(data_f,"durations.npy"),mmap_mode = mmap_mode)
    if max_n is not None:
        chunks = chunks[:max_n]
        paths = paths[:max_n]
        seqs = seqs[:max_n]
        seq_lens = seq_lens[:max_n]
        durations = durations[:max_n]
    if shuffle:
        perm = np.arange(len(chunks))
        perm = np.random.permutation(perm)
        chunks = chunks[perm]
        paths = paths[perm]
        seqs = seqs[perm]
        seq_lens = seq_lens[perm]
        durations = durations[perm]
    return chunks,paths,seqs,seq_lens,durations,delooped

def max_loop(seq,max_loop_size = 10):
    repeat_size = []
    for loop_size in np.arange(2,max_loop_size):
        curr_size = []
        for indent in np.arange(loop_size):
            curr_size.append(np.max([len(list(g)) for k,g in itertools.groupby(seq.replace("M","A")[indent:][::loop_size])]))
        repeat_size.append(np.min(curr_size))
    return np.max(repeat_size)

def m2a(kmer_i,idx2kmer,kmer2idx):
    kmer_i = int(kmer_i)
    if 'M' not in idx2kmer[kmer_i]:
        return kmer_i
    else:
        return kmer2idx[idx2kmer[kmer_i].replace('M','A')]

def a2m(kmer_i,idx2kmer,kmer2idx):
    kmer_i = int(kmer_i)
    if 'A' not in idx2kmer[kmer_i]:
        return kmer_i
    else:
        return kmer2idx[idx2kmer[kmer_i].replace('A','M')]

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--control","-c",required = True,type = str,help = "The control foldaer.")
    parser.add_argument("--modified","-m",required = True,type = str,help = "The modification folder.")
    parser.add_argument("--output","-o",required = True,type = str,help = "The output folder.")
    parser.add_argument("--sample_n","-n",type = int,default = 100000,help = "The number of sampling.")
    parser.add_argument("--mixture","-mix",type = float,default = 0.5,help = "The mixture ratio of the control and the modified, len(m)/len(c).")
    parser.add_argument("--exploration","-e",type = float,default = 0.1,help = "The exploration ratio.")
    parser.add_argument("--max",type = int,default = None,help = "The maximum number of segments to load.")
    parser.add_argument("--canonical_base","-cb",type = str,default = "A",help = "The canonical base.")
    parser.add_argument("--modified_base","-mb",type = str,default = "M",help = "The modified base.")
    parser.add_argument("--min_seq_len",'-ms',type = int,default = 7, help = "The minimum sequence length to be added in.")
    parser.add_argument("--min_seg_len", default = 500, type = int, help = "The minimum signal segment length to be added in.")
    args = parser.parse_args(sys.argv[1:])
    
    ##Testing code
    # np.random.seed(42)
    # home_f = os.path.expanduser("~")
    # scratch_f = home_f + "/bridge_scratch"
    # control_folder = scratch_f + "/Training_Datasets/EVA+NA12878IVT+ELIGOS+ONTBACTERIAMIX/control/"
    # modified_folder = scratch_f + "/Training_Datasets/EVA+NA12878IVT+ELIGOS+ONTBACTERIAMIX/modified/"
    # sample_n = 500000
    # max_n = None #For testing.
    # mix_ratio = 0.333 #M/A ratio
    # exploration = 0.2 #The decay of weight of edge after being samplinged.
    # out_f = home_f + "/Training_Datasets/cross_link_%dpct/"%(int(100*mix_ratio/(1.+mix_ratio)))

    ##Running code
    control_folder = args.control
    modified_folder = args.modified
    sample_n = args.sample_n
    mix_ratio = args.mixture
    exploration = args.exploration
    out_f = args.output
    max_n = args.max
    
    config = toml.load(os.path.join(control_folder,"config.toml"))
    config['exploration'] = exploration
    config['base_alternation']={args.modified_base:args.canonical_base}
    m2a_p = partial(m2a,idx2kmer = config['idx2kmer'],kmer2idx = config['kmer2idx_dict'])
    a2m_p = partial(a2m,idx2kmer = config['idx2kmer'],kmer2idx = config['kmer2idx_dict'])
    lg = LinkingGraph(config)
    lg.build_invariant_kmers([args.canonical_base,args.modified_base])
    print("Loading control data...")
    chunks_control,path_control,seqs_control,seq_lens_control,durations_control,delooped = load_data(control_folder,max_n = max_n,shuffle = True)
    if not delooped:
        results_dict = {"Fixed":0, "No need to fix":0, "Fix failed":0}
        print("The control data is not delooped. Deloopping the path.")
        with tqdm() as t:
         t.total = len(path_control)
         t.set_description("Fixing the path of control data.")
         for i,p in enumerate(path_control):
            s = seqs_control[i]
            if len(s) < args.min_seq_len:
                continue
            fixed_path,fixing = fixing_looping_path(p[:durations_control[i]],s,idx2kmer = config['idx2kmer'],modified_base=args.modified_base,canonical_base=args.canonical_base)
            if fixing is None:
                results_dict["Fix failed"] += 1
            elif fixing:
                results_dict['Fixed'] += 1
                path_control[i][:durations_control[i]] = fixed_path
            else:
                results_dict['No need to fix'] += 1
            t.set_description(f"Fixed {results_dict['Fixed']}, No need to fix {results_dict['No need to fix']}, Fix failed {results_dict['Fix failed']}")
            t.update()
        np.save(os.path.join(args.control,"path_fix.npy"),path_control)
    print("Loading modified data...")
    chunks_modified,path_modified,seqs_m6a,seq_lens_m6a,durations_m6a,delooped = load_data(modified_folder,max_n = max_n,shuffle = True)
    if not delooped:
        print("The modified data is not delooped. Deloopping the path.")
        results_dict = {"Fixed":0, "No need to fix":0, "Fix failed":0}
        with tqdm() as t:
            t.total = len(path_modified)
            t.set_description("Fixing the path of modified data.")
            for i,p in enumerate(path_modified):
                s = seqs_m6a[i]
                if len(s) < args.min_seq_len:
                    continue
                fixed_path,fixing = fixing_looping_path(p[:durations_m6a[i]],s,idx2kmer = config['idx2kmer'],modified_base=args.modified_base,canonical_base=args.canonical_base)
                if fixing is None:
                    results_dict["Fix failed"] += 1
                elif fixing:
                    results_dict['Fixed'] += 1
                    path_modified[i][:durations_m6a[i]] = fixed_path
                else:
                    results_dict['No need to fix'] += 1
                t.set_description(f"Fixed {results_dict['Fixed']}, No need to fix {results_dict['No need to fix']}, Fix failed {results_dict['Fix failed']}")
                t.update()
        np.save(os.path.join(args.modified,"path_fix.npy"),path_modified)

    if chunks_control.shape[0]*mix_ratio>chunks_modified.shape[0]:
        shrink = int(chunks_modified.shape[0]/mix_ratio)
        chunks_control = chunks_control[:shrink]    
        path_control = path_control[:shrink]
        seqs_control = seqs_control[:shrink]
        seq_lens_control = seq_lens_control[:shrink]
    else:
        shrink = int(chunks_control.shape[0]*mix_ratio)
        chunks_modified = chunks_modified[:shrink]
        path_modified = path_modified[:shrink]
        seqs_m6a = seqs_m6a[:shrink]
        seq_lens_m6a = seq_lens_m6a[:shrink]
    print("Add control data (%d chunks) into the graph."%(chunks_control.shape[0]))
    lg.add_reads(chunks_control,path_control,durations_control,min_seg_len = args.min_seg_len,transition = m2a_p)
    print("Add modified data (%d chunks) into the graph."%(chunks_modified.shape[0]))
    lg.add_reads(chunks_modified,path_modified,durations_m6a,min_seg_len = args.min_seg_len,transition = a2m_p)
    print("Sampling the data.")
    s,p,p_i = lg.sampling(chunks_control.shape[1],sample_n)
    seqs = np.asarray([kmers2seq(x[x!=-1],lg.idx2kmer) for x in tqdm(p,total = len(p),desc = "Transfer to sequence")])
    seqs_len = np.asarray([len(x) for x in seqs])
    # This is no longer needed as we already fix the loop in deloop function.
    # loop_filter = np.asarray([max_loop(seq)<4 for seq in tqdm(seqs,total = len(seqs),desc = "apply filter")])
    # s,p,seqs,seqs_len = s[loop_filter],p[loop_filter],seqs[loop_filter],seqs_len[loop_filter]
    print("%d samples left after filtering"%(len(s)))
    print("Maximum sequence length %d"%(np.max(seqs_len)))
    print("Saving the data")
    os.makedirs(out_f,exist_ok = True)
    np.save(os.path.join(out_f,"chunks.npy"),s)
    np.save(os.path.join(out_f,"path.npy"),p)
    np.save(os.path.join(out_f,"seqs"),seqs)
    np.save(os.path.join(out_f,"seq_lens.npy"),seqs_len)
    np.save(os.path.join(out_f,"path_idx.npy"),p_i)
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 17 18:21:00 2022

@author: haotian teng
"""
import os
import copy
import toml
import torch
import itertools
import torchvision
import numpy as np
import torch.utils.data as data
from xron.utils.seq_op import findall
from typing import List,Dict,Callable

### Test module ###
from time import time

class Kmer_Dataset(data.Dataset):
    def __init__(self,
                 chunks:np.array,
                 durations:np.array,
                 kmer_labels:np.array,
                 labels:np.array = None,
                 transform:torchvision.transforms.transforms.Compose=None):
        
        self.chunks = chunks
        self.duration = durations.astype(np.int64)
        self.labels = kmer_labels
        self.transform = transform
        self.y = labels
                
    def __len__(self):
        return len(self.chunks)

    def __getitem__(self, idx):
        sample = {'signal': self.chunks[idx][:,None].astype(np.float32), 
                  'duration': self.duration[idx],
                  'labels': self.labels[idx]}
        if self.y is not None:
            sample['y'] = self.y[idx]
        if self.transform:
            sample = self.transform(sample)
        return sample
    
def flatten(nested_list:List[List])->List:
    """
    Flat a nested list

    Parameters
    ----------
    nested_list : List[List]
        A nested list.

    Returns
    -------
    List
        The flatten list.

    """
    return [x for sub_list in nested_list for x in sub_list]

class Normalizer(object):
    def __init__(self,
                 use_dwell:bool = True,
                 statistics:Callable = None,
                 no_scale:bool = False,
                 min_dwell:int = 5,
                 effective_kmers:List[int] = None):
        self.use_dwell = use_dwell
        self.statistics = np.median if statistics is None else statistics
        self.no_scale = no_scale
        self.min_dwell = min_dwell
        self.effective_kmers = effective_kmers
    def __call__(self,
                 *args,
                 **kwargs):
        if self.use_dwell:
            return self.rescale_dwell(*args, **kwargs)
        else:
            return self.rescale(*args, **kwargs)
        
    
    def dwell(self,
              sig:np.array,
              path:np.array,
              duration:int):
        """
        Return the dwell property of the signal
    
        Parameters
        ----------
        sig : np.array, shape [N].
            The original signal.
        path : np.array, shape [N]
            The decoded path of the current signal.
        duration : int
            The duration of the signal. The default is None.
        statistics: func, optional, default is median.
            The statistics of the dwell to return.
        min_dwell: int, optional, default is 3.
            The minimum length of the dwell to be considered.
        Returns
        -------
        TYPE
            Return the chosen statistics of the dwells.
    
        """
        gp = [list(y) for x,y  in itertools.groupby(zip(sig[:duration],path[:duration]),key = lambda x: x[1])]
        if self.effective_kmers is None:
            dwells = [[y[0] for y in x] for x in gp if len(x)>=self.min_dwell]
        else:
            dwells = [[y[0] for y in x] for x in gp if len(x)>=self.min_dwell and x[0][1] in self.effective_kmers]
            if len(dwells) < 5:
                # print("Warning, there are very few dwells obtained from effective kmers, whole kmers list will be used instead.")
                # dwells = [[y[0] for y in x] for x in gp if len(x)>=self.min_dwell]
                dwells = [[0.]]
        return [self.statistics(x) for x in dwells]
        
    def rescale(self,
                sig:np.array,
                rc_sig:np.array,
                duration:np.array,
                return_sig:bool = True):
        """
        Rescale the input signal using the reconstructed signal.
    
        Parameters
        ----------
        sig : np.array, shape [B,N].
            The original signal.
        rc_sig : np.array, shape [B,N]
            The reconstructed signal.
        duration : np.array, optional, shape [B]
            The duration of the signal batch. The default is None.
        return_sig: bool, optional, default is True.
            If return the normalized signal or scale.
        Returns
        -------
        TYPE
            Return the rescaled signal.
    
        """
        B,L = sig.shape
        if sig.ndim == 1:
            sig = sig[None,:]
        if rc_sig.ndim == 1:
            rc_sig = rc_sig[None,:]
        if duration is None:
            N = np.asarray([len(sig)]*sig.shape[0])
        else:
            if type(duration) == type(1):
                duration = np.asarray([duration])
            N = duration
        if self.no_scale:
            scale = np.asarray([1])
        else:
            scale = (N*np.sum(sig*rc_sig,axis = 1) - np.sum(sig,axis = 1)*np.sum(rc_sig,axis = 1))/(N*np.sum(sig**2,axis = 1)-np.sum(sig,axis = 1)**2)
        for i,s in enumerate(rc_sig):
            s[duration[i]:] = 0
        shift = np.mean(rc_sig,axis = 1)*L/duration-scale*np.mean(sig,axis = 1)*L/duration
        if return_sig:
            return scale[:,None]*sig + shift[:,None]
        else:
            return scale, shift
    
    def rescale_dwell(self,
                      sig:np.array,
                      rc_sig:np.array,
                      duration:np.array,
                      path:np.array):
        """https://researchcomputing.princeton.edu/support/knowledge-base/singularity
        Rescale the input signal using the reconstructed signal.
    
        Parameters
        ----------
        sig : np.array, shape [B,N].
            The original signal.
        rc_sig : np.array, shape [B,N]
            The reconstructed signal.
        duration : np.array, shape [B]
            The duration of the signal batch. The default is None.
        path : np.array, shape [B,N]
            The decoded path of the signal.
        Returns
        -------
        TYPE
            Return the rescaled signal.
    
        """
        dwell_sig = [self.dwell(s,p,d) for s,p,d in zip(sig,path,duration)]
        dwell_rc = [self.dwell(s,p,d) for s,p,d in zip(rc_sig,path,duration)]
        ls = [len(x) for x in dwell_sig]
        max_len = max(ls)
        dwell_sig_pad = [x + [0]*(max_len - len(x)) for x in dwell_sig]
        dwell_rc_pad = [x + [0]*(max_len - len(x)) for x in dwell_rc]
        scale,shift = self.rescale(np.array(dwell_sig_pad),
                                   np.array(dwell_rc_pad),
                                   np.array(ls),
                                   return_sig = False)
        return scale[:,None]*sig + shift[:,None]
    

class Kmer2Transition(object):
    """
    Convert string sequence to integer index array
    
    Parameters
    ----------
    alphabeta : str
        The base alphabeta combination.
    k: int
        The size of the knmer.
    T_max: int
        The maximum length of each sequence.
    kmer2idx: dict
        The dictionary link the kmer to the index.
    idx2kmer: List
        The list give the kmer.
    neighbour_kmer: int
        The number of neighbourhood kmer to be considered when building the 
        transition matrix, if 0 then will only consider the current kmer when
        building transition matrix.
    base_alternation: dict, optional
        The dictionary gives the methylation replacement of the base, e.g. 
        {'A':'MX'}, means base A can be two kinds of methylation bases M and X.
        Default is None, no replacement will be made.
    base_prior: Dict, optional
        The prior probability of each base, this is useful when fitting in-vitro
        transcription dataset, default is None, that every base has equal prior
        (1).
    kmer_replacement: bool, optional
        If we consider kmer replacement during transferring to the transition
        matrix.
            
    """
    def __init__(self,alphabeta:str, 
                 k:int, 
                 T_max:int, 
                 kmer2idx:dict, 
                 idx2kmer:List, 
                 neighbour_kmer:int = 2, 
                 base_alternation:Dict[str, str] = None,
                 base_prior = None,
                 kmer_replacement = False,
                 out_format = "sparse"):
        self.n_base = len(alphabeta)
        self.alphabeta = alphabeta
        self.k = k
        self.n_states = self.n_base**self.k
        self.T_max = T_max
        self.neighbour = neighbour_kmer
        self.kmer2idx = kmer2idx
        self.idx2kmer = idx2kmer
        self.base_alternation = base_alternation
        if base_prior is None:
            self.base_prior = {x:1 for x in self.alphabeta}
        else:
            self.base_prior = base_prior
            for b in self.alphabeta:
                if b not in self.base_prior.keys():
                    self.base_prior[b] = 1
        self._build_background_matrix()
        self._build_methylation_alternation()
        self.kmer_replacement = kmer_replacement
        self.out_format = out_format
        
    def _build_background_matrix(self):
        """
        Build the background transition matrix, i.e AACAA -> ACAAA,ACAAC,ACAAG,
        ACAAT,...

        """
        bg_source, bg_target = [],[]
        bg_idx = []
        self.from_idx = np.zeros((self.n_states,self.n_base+1))
        bg_vals = []
        for i in np.arange(self.n_states):
            kmer = self.idx2kmer[i]
            bg_source.append(i)
            bg_target.append(i)
            bg_vals.append(1)
            curr_tgt = [i]
            for base in self.alphabeta:
                new_kmer = kmer[1:]+base
                j = self.kmer2idx[new_kmer]
                if i==j:
                    curr_tgt.append(j)
                    continue
                bg_source.append(i)
                bg_target.append(j)
                bg_vals.append(self.base_prior[base])
                curr_tgt.append(j)
            bg_idx.append(curr_tgt)
        self.bg = np.zeros((self.n_states,self.n_states),dtype = np.float32)
        self.bg[bg_source,bg_target] = np.array(bg_vals)
        self.to_idx = np.array(bg_idx)
        for i in np.arange(self.n_states):
            kmer = self.idx2kmer[i]
            self.from_idx[i][0] = i
            for j,base in enumerate(self.alphabeta):
                new_kmer = base + kmer[:-1]
                new_idx = self.kmer2idx[new_kmer]
                self.from_idx[i][j+1] = new_idx
        self.idx_map = [self.from_idx,self.to_idx]
        
    def _build_methylation_alternation(self):
        """
        Create the methylation kmer alternative dictionary, e.g. ACGMT:[ACGMT,
        ACGAT, MCGAT, MCGMT], all kmer is represented using its index.

        """
        self.kmer_alternation = {}
        for i in np.arange(self.n_states):
            kmer = self.idx2kmer[i]
            alternations = []
            for perm_kmer in self._alternative_kmer(kmer):
                alternations.append(self.kmer2idx[perm_kmer])
            self.kmer_alternation[i] = alternations
            
    def _alternative_kmer(self,kmer):
        """
        Return an iterator gives the alternative kmer of the current kmer, e.g.
        given the current kmer is 'AACCG', the alternative kmer list will be
        ['MMCCG','MACCG','AMCCG','AACCG'], all kmer will be transferred to its
        corresponding indexs.

        Parameters
        ----------
        kmer : str
            The current kmer.

        Yields
        ------
        str
            Iterate over all the alternative kmer (include the kmer itself).

        """
        for k,v in self.base_alternation.items():
            for x in v:
                kmer = kmer.replace(x,k)
            v += k
            locs = findall(kmer,k)
            for x in itertools.product(v,repeat = len(locs)):
                curr_kmer = [x for x in kmer]
                for i,l in enumerate(locs):
                    curr_kmer[l] = x[i]
                yield ''.join(curr_kmer)
    
    def legal_transition(self,kmer_list:List[int]) -> List[set]:
        """
        Return the legal transition term given a kmer list.

        Parameters
        ----------
        kmer_list : List[int]
            DESCRIPTION.

        Returns
        -------
        List[set]
            DESCRIPTION.

        """
    
    def kmer2transition(self,kmer_seq):
        transitions = []
        # kmer_seq = copy.deepcopy(kmer_seq) #To avoid torch.dataLoader blowing up shared memory according to this issue:https://github.com/pytorch/pytorch/issues/11201
        condensation = [(k,len(list(g))) if k>=0 else (0,len(list(g))) for k,g in itertools.groupby(kmer_seq)]
        condensed_kmer_seqs = [x[0] for x in condensation]
        condensed_transitions = self._get_transition_pairs(condensed_kmer_seqs)
        for i,(kmer,period) in enumerate(condensation):
            curr = flatten(condensed_transitions[max(0,i-self.neighbour):i+self.neighbour+1])
            source,target = list(zip(*curr))
            curr_transition = torch.sparse_coo_tensor([source,target], self.bg[source,target],(self.n_states,self.n_states))
            transitions += [curr_transition]*period
        return transitions
    
    def kmer2compact_transition(self,kmer_seq):
        transitions = []
        condensation = [(k,len(list(g))) if k>=0 else (0,len(list(g))) for k,g in itertools.groupby(kmer_seq)]
        condensed_kmer_seqs = [x[0] for x in condensation]
        condensed_transitions = self._get_transition_pairs(condensed_kmer_seqs)
        for i,(kmer,period) in enumerate(condensation):
            curr = flatten(condensed_transitions[max(0,i-self.neighbour):i+self.neighbour+1])
            transition = self.compact_transition_tensor(curr, [self.bg[s,t] for s,t in curr])
            transitions += [transition]*period
        return transitions
    
    def compact_transition_tensor(self,src_tgt:List,value:List) -> torch.tensor:
        """
        Generate a compact transition tensor given the soruce-target paris and
        the values.

        Parameters
        ----------
        src_tgt : List with shape [L]
            A List contains the source and target tuple.
        value : List with shape [L]
            List contains the values, this list should have the same length as
            the source and target.

        Returns
        -------
        A compact torch tensor with shape [N,B+1], where N is the number of kmers
        and B is the number of bases, and 0 index always stand for kmers stay.

        """
        trs = torch.zeros((self.n_states,self.n_base + 1),dtype = torch.float)
        trt = torch.zeros((self.n_base + 1,self.n_states),dtype = torch.float)
        for (src,tgt),val in zip(src_tgt,value):
            trs[src][np.where(self.to_idx[src] == tgt)[0][0]] += val
        for (src,tgt),val in zip(src_tgt,value):
            trt[np.where(self.from_idx[tgt]==src)[0][0]][tgt] += val
        return torch.cat((trs,trt.T),dim = 1)
        
    
    def _get_transition_pairs(self,condensed_kmer_seqs):
        if self.kmer_replacement:
            condensed_kmer_seqs = [self.kmer_alternation[x] for x in condensed_kmer_seqs]
        else:
            condensed_kmer_seqs = [[x] for x in condensed_kmer_seqs]
        transitions = []
        for i in np.arange(0,len(condensed_kmer_seqs)):
            if i == 0:
                curr_trans = list(zip(condensed_kmer_seqs[i],condensed_kmer_seqs[i]))
            else:
                curr_trans = list(itertools.product(condensed_kmer_seqs[i-1],condensed_kmer_seqs[i]))+ list(zip(condensed_kmer_seqs[i],condensed_kmer_seqs[i]))
            curr_trans = set([x for x in curr_trans if self.bg[x[0],x[1]]])
            transitions.append(curr_trans)
        return transitions
    
    def __call__(self,sample:Dict):
        func = self.kmer2transition if self.out_format == "sparse" else self.kmer2compact_transition
        transformed = {key:(value if key != 'labels' else func(value)) for key, value in sample.items()}
        transformed['kmers'] = sample['labels']
        return transformed
        

if __name__ == "__main__":
    from torchvision import transforms
    torch.multiprocessing.set_sharing_strategy('file_system')
    import time
    # kmers_f = "/home/heavens/BRIDGE_SCRATCH/NA12878_RNA_IVT/xron_output/extracted_kmers"
    kmers_f = "/home/haotian/bridge_scratch/NA12878_RNA_IVT/xron_partial/extracted_kmers"
    chunks = np.load(os.path.join(kmers_f,"chunks.npy"))
    durations = np.load(os.path.join(kmers_f,"durations.npy"))
    kmers = np.load(os.path.join(kmers_f,"kmers.npy"))
    config = toml.load(os.path.join(kmers_f,"config.toml"))
    
    ### Test Kmer2Transition class
    k2t = Kmer2Transition('ACGTM',5,4000,config['kmer2idx_dict'],config['idx2kmer'],base_alternation = {"A":"M"}, kmer_replacement = True, out_format = "compact")
    t = k2t.kmer2transition(kmers[0])
    
    ### Test data loader
    dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
    loader = data.DataLoader(dataset,batch_size = 2, shuffle = True,num_workers=2)
    start = time.time()
    for i_batch,batch in enumerate(loader):
        print("Batch loading time %.2f"%(time.time() - start))
        time.sleep(3)
        start = time.time()
        # print(batch['signal'].shape)
        # print(batch['duration'].shape)
        # print(batch['labels'])



from setuptools import setup,find_packages,Extension
from setuptools.command.install import install
import numpy as np
import os
# read the contents of your README file
with open('README.md') as f:
    long_description = f.read()
print(long_description)
class CustomInstallCommand(install):
    def run(self):
        print("\nThis package is licensed under the GNU General Public License v3.0 (GPLv3).")
        print("Please refer to the LICENSE file for more information.\n")
        install.run(self)

install_requires=[
  'h5py',
  'mappy>=2.10.0',
  'numpy>=1.13.3',
  'statsmodels>=0.8.0',
  'tqdm>=4.23.0',
  'scipy>=1.0.1',
  'biopython==1.73',
  'google-auth==2.18.1',
  'oauthlib==3.2.2',
  'packaging>=18.0',
  'ont-fast5-api>=0.3.1',
  'wget>=3.2',
  'pysam>=0.21.0',
  'tensorboard',
  'matplotlib',
  'seaborn',
  'pandas',
  'toml',
  'fast-ctc-decode',
  'editdistance>=0.5.3',
  'torch>=1.12.0',
  'torchvision>=0.13.0',
  'torchaudio>=0.12.0',
  'boostnano',
]
exec(open('xron/_version.py').read()) #readount the __version__ variable
setup(
  name = 'xron',
  packages = find_packages(exclude=["*.test", "*test.*", "test.*", "test"]),
  version = __version__,
  include_package_data=True,
  description = 'A deep neural network basecaller for nanopore sequencing.',
  author = 'Haotian Teng',
  author_email = 'havens.teng@gmail.com',
  url = 'https://github.com/haotianteng/Xron', 
  download_url = 'https://github.com/haotianteng/Xron/archive/1.0.0.tar.gz', 
  keywords = ['basecaller', 'nanopore', 'sequencing','neural network','RNA methylation'], 
  license="GPL 3.0",
  classifiers = ['License :: OSI Approved :: GNU General Public License v3 (GPLv3)'],
  install_requires=install_requires,
  entry_points={'console_scripts':['xron=xron.entry:main'],},
  long_description=long_description,
  include_dirs = [np.get_include()],
  long_description_content_type='text/markdown',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  3 11:47:48 2022

@author: heavens
"""
import os
import h5py
import argparse
from xron.utils.seq_op import fast5_shallow_iter
from xron.utils.fastIO import read_fastq

def indexing(args):
    fastq_records = read_fastq(args.fastq)
    fast5_records = {}
    print("Indexing fastq files.")
    with open(args.fastq+'.index', 'w+') as f:
        for root,abs_path in fast5_shallow_iter(args.fast5,tqdm_bar = True):
            read_ids = list(root.keys())
            for id in read_ids:
                fast5_records[id[5:]] = abs_path #read_id is like "read_00000000-0000-0000-0000-0000000000"
        for id in fastq_records['name']:
            if id in fast5_records.keys():
                f.write(id+'\t'+fast5_records[id]+'\n')
            else:
                raise KeyError('fastq readid %s not found in fast5'%(id))
    print("Indexing fastq file has been finished.")            
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('--fast5', required = True, type=str, help='folder that contains fast5 output')
    parser.add_argument('--fastq', required = True, type=str, help='The merged fastq file')
    args = parser.parse_args()
    indexing(args)


# Copyright 2017 The Chiron Authors. All Rights Reserved.
#
#This Source Code Form is subject to the terms of the Mozilla Public
#License, v. 2.0. If a copy of the MPL was not distributed with this
#file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Initial file for the package"""
from xron._version import __version__


"""
Created on Mon Mar 28 23:54:38 2022

@author: Haotian Teng
"""
import h5py
import numpy as np
from matplotlib import pyplot as plt
strides = 11
root = h5py.File("/home/heavens/Documents/UCSC_Run1_20180129_IVT_RNA_201.fast5",'r')
for read_id in root:
    if read_id == "read_40634cc4-ff9c-41f3-bd2c-901cb23c0179":
        plt.figure(figsize = (40,20))
        plt.plot(np.asarray(root[read_id]['Raw/Signal'])[::-1])
        read_h = root[read_id]
        basecalled = read_h['Analyses/Basecall_1D_000/BaseCalled_template']
        hidden = np.asarray(basecalled['Hidden'])
        signal = np.asarray(read_h['Raw/Signal'])
        move = basecalled['Move']
        position_sig = np.repeat(np.cumsum(move)-1,repeats = strides).astype(np.int32)[:len(signal)]
        segmentation = np.where(np.diff(position_sig))[0]
        
        plt.figure(figsize = (80,40))
        max_T = 40000
        plt.plot(signal[::-1][-(max_T+5000):],color = "red")
        plt.vlines(x = segmentation[segmentation>(segmentation[-1]-max_T)] - len(signal)+max_T, 
                   ymin = min(signal), 
                   ymax = max(signal), 
                   color = "green", 
                   linestyles = "dotted", 
                   linewidth = 0.01, 
                   antialiased=False)
        
        plt.figure(figsize = (80,40))
        max_T = 40000
        plt.plot(signal[:max_T],color = "red")
        segmentation = segmentation[-1] - segmentation[::-1]
        plt.vlines(x = segmentation[segmentation<max_T], 
                   ymin = min(signal), 
                   ymax = max(signal), 
                   color = "green", 
                   linestyles = "dotted", 
                   linewidth = 0.01, 
                   antialiased=False)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul  1 04:38:02 2022

@author: haotian teng
"""
import torch
import itertools
import numpy as np

def log_domain_sparse_matmul(A:torch.sparse_coo_tensor, log_B:torch.Tensor, dim:int = -1):
    """
    Do a sparse-dense tensor multiplication and reduced on the given dimension.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.
    dim : int, optional
        The dimension to perform reduction on A. The default is -1.

    Returns
    -------
    A sparse tensor.

    """
    idxs,update = log_domain_sparse_product(A,log_B)
    shape_A = torch.tensor(A.shape)
    remain_dims = np.delete(np.arange(idxs.shape[0]),dim)
    remain_idxs = idxs[remain_dims,:].T.tolist()
    key_func = lambda x: x[1]
    update = sorted(zip(update,remain_idxs),key = key_func)
    nested = [ (k,list(g)) for k,g in itertools.groupby(update,key = key_func)]
    nested_vals = [[y[0] for y in x[1]] for x in nested]
    nested_idxs = [x[0] for x in nested]
    max_cols = max([len(x) for x in nested_vals])
    padded = torch.tensor([x + [-np.inf]*(max_cols - len(x)) for x in nested_vals],device = A.device)
    return torch.sparse_coo_tensor(indices = list(zip(*nested_idxs)),
                                   values = torch.logsumexp(padded,dim = 1),
                                   size = shape_A[remain_dims].tolist(),
                                   device = A.device)

def log_domain_sparse_matmul_new(A:torch.sparse_coo_tensor, log_B:torch.Tensor, dim:int = -1):
    """
    Do a sparse-dense tensor multiplication and reduced on the given dimension.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.
    dim : int, optional
        The dimension to perform reduction on A. The default is -1.

    Returns
    -------
    A sparse tensor.

    """
    idxs,update = log_domain_sparse_product(A,log_B)
    shape_A = torch.tensor(A.shape)
    remain_dims = np.delete(np.arange(idxs.shape[0]),dim)
    remain_idxs = idxs[remain_dims,:].T
    uniq,uniq_idxs = torch.unique(remain_idxs,dim = 0, return_inverse = True)
    nested_vals = [update[uniq_idxs == i].tolist() for i in np.arange(len(uniq))]
    # nested_vals = [1]*len(uniq)
    max_cols = max([len(x) for x in nested_vals])
    padded = torch.tensor([x + [-np.inf]*(max_cols - len(x)) for x in nested_vals],device = A.device)
    return torch.sparse_coo_tensor(indices = uniq.T,
                                   values = torch.logsumexp(padded,dim = 1),
                                   size = shape_A[remain_dims].tolist(),
                                   device = A.device)


def log_domain_sparse_max(A:torch.sparse_coo_tensor, log_B:torch.Tensor, dim:int = -1):
    """
    Do a sparse-dense tensor multiplication and reduced on the given dimension.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.
    dim : int, optional
        The dimension to perform reduction on A. The default is -1.

    Returns
    -------
    A sparse tensor.

    """
    idxs,update = log_domain_sparse_product(A,log_B)
    n_dims_A = idxs.shape[0]
    shape_A = torch.tensor(A.shape)
    remain_dims =np.delete(np.arange(n_dims_A),dim)
    remain_idxs = list(zip(*[idxs[x].tolist() for x in remain_dims]))
    reduced_idxs = idxs[dim].tolist()
    update = update.tolist()
    key_func = lambda x: x[2]
    update = sorted(zip(update,reduced_idxs,remain_idxs),key = key_func)
    nested = [ (k,list(g)) for k,g in itertools.groupby(update,key = key_func)]
    nested_vals = [[y[0] for y in x[1]] for x in nested]
    nested_idxs = list(zip(*[x[0] for x in nested]))
    nested_reduced_idxs = [[y[1] for y in x[1]] for x in nested]
    max_cols = max([len(x) for x in nested_vals])
    padded = torch.tensor([x + [-np.inf]*(max_cols - len(x)) for x in nested_vals],device = A.device)
    padded_reduced_idxs = torch.tensor([x + [-1]*(max_cols - len(x)) for x in nested_reduced_idxs],device = A.device,dtype = torch.long)
    result,argmax = torch.max(padded,dim = 1)
    argmax = torch.gather(padded_reduced_idxs,dim = 1,index = argmax[:,None]).squeeze(dim = 1)
    max_idx = torch.zeros(shape_A[remain_dims].tolist(),device = A.device,dtype = torch.long)
    max_idx[tuple(nested_idxs)] = argmax
    return torch.sparse_coo_tensor(indices = nested_idxs,
                                   values = result,
                                   size = shape_A[remain_dims].tolist(),
                                   device = A.device).to_dense(),max_idx


def log_domain_sparse_product(A:torch.sparse_coo_tensor, log_B:torch.Tensor):
    """
    Do a sparse-dense tensor production.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.

    Returns
    -------
    idxs,vals

    """
    A = A.coalesce()
    idxs = A.indices()
    log_vals = A.values().log()
    shape_A = torch.tensor(A.shape)
    shape_B = torch.tensor(log_B.shape)
    n_dims_A = idxs.shape[0]
    n_dims_B = len(shape_B)
    assert n_dims_A == n_dims_B, "Tensor has different number of dimensions."
    assert torch.all(torch.logical_or(shape_A == shape_B,shape_B == 1)), "Shape mismatch, got {} and {}. Broadcast only supported on dense tensor.".format(shape_A,shape_B)
    idxs_B = idxs.clone()
    idxs_B[torch.where(shape_B==1)] = 0
    update = log_B[tuple(idxs_B)] + log_vals
    return idxs,update

if __name__ == "__main__":
    from time import time
    N = 3000
    K = 3125
    B = 100
    i = torch.tensor([np.random.randint(low = 0, high = B, size = N),
                      np.random.randint(low = 0, high = K, size = N),
                      np.random.randint(low = 0, high = K, size = N)])
    v = torch.tensor(np.random.rand(N), dtype=torch.float32)
    s = torch.sparse_coo_tensor(i, v, [B, K, K])
    d = torch.rand((B,K,1))
    start = time()
    result = log_domain_sparse_matmul(s,d,dim = 1)
    print(time()-start)
    start = time()
    result_new = log_domain_sparse_matmul_new(s,d,dim=1)
    print(start-time())
    
    idxs = s.coalesce().indices()
    remain_dims = np.delete(np.arange(idxs.shape[0]),1)
    remain_idxs = idxs[remain_dims].T


"""
Created on Sun Dec 26 19:38:25 2021

@author: Haotian Teng
"""
import os
import sys
import argparse
from xron_model import MM_CONFIG

def main(args):
    base_config = MM_CONFIG()
    grid_key = [base_config.CNN['Layers'][0]['out_channels'],
                base_config.CNN['Layers'][1]['out_channels'],
                base_config.CNN['Layers'][2]['out_channels'],
                base_config.RNN['hidden_size'],
                base_config.PORE_MODEL['K']]
    grid = [[4,16,64],
            [16,64,128],
            [256,512,768],
            [256,512,768],
            [2,3,4,5]]
    

# if __name__ == "__main__":
#     parser = argparse.ArgumentParser(
#         description='Calling training module.')
#     parser.add_argument('--module', required = True,
#                         help = "The training module to call, can be Embedding, Supervised and Reinforce")
#     parser.add_argument('-o', '--model_folder', required = True,
#                         help = "The folder to save folder at.")
#     args = parser.parse_args(sys.argv[1:])
#     os.makedirs(args.model_folder,exist_ok=True)
#     main(args)